# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Digest`."""
import hashlib
from typing import Any

from .data import Data


class Digest(Data):
    """A :class:`~ckms.core.types.Data` implementation that represents a
    digest.
    """
    __module__: str = 'ckms.types'
    digest_oid: dict[str, str] = {
        '1.3.14.3.2.26'         : 'sha1',
        '2.16.840.1.101.3.4.2.1': 'sha256',
        '2.16.840.1.101.3.4.2.2': 'sha384',
        '2.16.840.1.101.3.4.2.3': 'sha512',
    }
    digest_size: dict[str, int] = {
        'sha256': 32,
        'sha384': 48,
        'sha512': 64,
        'sha3_256': 32,
        'sha3_384': 48,
        'sha3_512': 64,
    }

    @staticmethod
    def hasher(oid: str) -> Any:
        return hashlib.new(Digest.digest_oid[oid])

    @classmethod
    def frombytes(cls, algorithm: str, buf: bytes) -> 'Digest':
        hasher = cls.hashers[algorithm]()
        hasher.update(buf)
        return cls(buf=hasher.digest(), digest=algorithm)

    @classmethod
    def fromoid(cls, data: bytes, oid: str) -> 'Digest':
        """Return a new :class:`Digest` configured by OID."""
        return cls.frombytes(algorithm=cls.digest_oid[oid], buf=data)

    def __init__(self, *, buf: bytes, digest: str, **kwargs: Any):
        if digest not in self.hashers:
            raise ValueError(f'Unknown digest algorithm: {digest}.')
        if len(buf) != self.digest_size[digest]:
            raise ValueError(f'Invalid digest length for {digest}.')
        super().__init__(buf=buf, digest=digest, **kwargs)

    async def digest(
        self,
        algorithm: str | None = None,
        length: int | None = None,
        encoding: str | None = None
    ) -> bytes:
        if algorithm is not None and algorithm != self._algorithm:
            raise ValueError("Requested algorithm differs from digest algorithm.")

        # No async call here since the data is already in memory.
        return self.encode(bytes(self), encoding=encoding)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ipaddress
from typing import Literal

from pyasn1.type import univ

from .name import Name


class IPAddressName(Name):
    kind: Literal['iPAddress'] = 'iPAddress'
    value: ipaddress.IPv4Address | ipaddress.IPv6Address

    @classmethod
    def parse_value(
        cls,
        value: bytes | str | univ.OctetString
    ) -> ipaddress.IPv4Address | ipaddress.IPv6Address:
        if isinstance(value, univ.OctetString):
            value = bytes(value)
        return ipaddress.ip_address(value)

    def encode_value(self) -> bytes:
        return self.value.packed



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`EllipticCurveOperation`."""
from typing import Any

from .operation import Operation


class EllipticCurveOperation(Operation):
    """Represents a cryptographic operation using an elliptic curve key."""
    __module__: str = 'ckms.core.types'
    _digest: str
    _signature: bytes | None

    @property
    def digest(self) -> str:
        return self._digest

    @property
    def signature(self) -> bytes | None:
        return self._signature

    def __init__(
        self,
        *,
        digest: str,
        signature: bytes | None = None,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        self._digest = digest
        self._signature = signature

    async def get_digest(self) -> bytes:
        """Return a digest of the message."""
        return await self.content.digest(algorithm=self.digest)


# pylint: skip-file
from typing import cast
from typing import Any

import pytest

from ckms import core
from ckms.core.local import LocalProvider
from ckms.core.random import RandomProvider


RSA_SIGNING_ALGORITHMS: list[str] = [
    'RS256',
    'RS384',
    'RS512',
    'PS256',
    'PS384',
    'PS512',
]

RSA_ENCRYPTION_ALGORITHMS: list[str] = [
    'RSA-OAEP',
    'RSA-OAEP-256',
    'RSA-OAEP-384',
    'RSA-OAEP-512',
]

EC_SIGNING_ALGORITHMS: list[str] = [
    'ES256',
    'ES384',
    'ES512',
    'ES256K',
]

ECDH_ENCRYPTION_ALGORITHMS: list[str] = [
    'ECDH-ES',
    'ECDH-ES+A128KW',
    'ECDH-ES+A192KW',
    'ECDH-ES+A256KW',
]

SYMMETRIC_ENCRYPTION_ALGORITHMS: list[str] = [
    'A128KW',
    'A192KW',
    'A256KW',
    'A128GCMKW',
    'A192GCMKW',
    'A256GCMKW',
]

HMAC_SIGNING_ALGORITHMS: list[str] = ['HS256', 'HS384', 'HS512']

OKP_SIGNING_CURVES: list[str] = ['Ed448', 'Ed25519']

OKP_ENCRYPTION_CURVES: list[str] = ['X448', 'X25519']

SUPPORTED_KEYS: list[dict[str, Any]] = [
    {
        'key': {'path': 'pki/rsa.key'}
    }
]


@pytest.fixture
def provider() -> LocalProvider:
    return cast(LocalProvider, core.provider('local'))


@pytest.mark.parametrize("algorithm", RSA_SIGNING_ALGORITHMS)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_rsa_signing(
    provider: LocalProvider,
    algorithm: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'RSA',
        'algorithm': algorithm,
        'provider': 'local'
    })
    assert spec.kty == 'RSA'
    assert spec.use == 'sig'
    assert spec.algorithm == algorithm


@pytest.mark.parametrize("algorithm", RSA_ENCRYPTION_ALGORITHMS)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_rsa_encryption(
    provider: LocalProvider,
    algorithm: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'RSA',
        'algorithm': algorithm,
        'provider': 'local'
    })
    assert spec.kty == 'RSA'
    assert spec.use == 'enc'
    assert spec.algorithm == algorithm


@pytest.mark.parametrize("algorithm", EC_SIGNING_ALGORITHMS)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_ec_signing(
    provider: LocalProvider,
    algorithm: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'EC',
        'algorithm': algorithm,
        'provider': 'local'
    })
    assert spec.kty == 'EC'
    assert spec.use == 'sig'
    assert spec.algorithm == algorithm


@pytest.mark.parametrize("algorithm", HMAC_SIGNING_ALGORITHMS)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_symmetric_sign(
    provider: LocalProvider,
    algorithm: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'oct',
        'algorithm': algorithm,
        'provider': 'local'
    })
    assert spec.kty == 'oct'
    assert spec.use == 'sig'
    assert spec.algorithm == algorithm


@pytest.mark.parametrize("algorithm", SYMMETRIC_ENCRYPTION_ALGORITHMS)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_symmetric_encrypt(
    provider: LocalProvider,
    algorithm: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'oct',
        'algorithm': algorithm,
        'provider': 'local'
    })
    assert spec.kty == 'oct'
    assert spec.use == 'enc'
    assert spec.algorithm == algorithm


@pytest.mark.parametrize("curve", OKP_SIGNING_CURVES)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_okp_signing(
    provider: LocalProvider,
    curve: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'OKP',
        'curve': curve,
        'provider': 'local'
    })
    assert spec.kty == 'OKP'
    assert spec.use == 'sig'
    assert spec.algorithm == "EdDSA"
    assert spec.curve == curve


@pytest.mark.parametrize("curve", OKP_ENCRYPTION_CURVES)
@pytest.mark.parametrize("params", SUPPORTED_KEYS)
def test_validate_spec_okp_encryption(
    provider: LocalProvider,
    curve: str,
    params: dict[str, Any]
):
    spec = provider.parse_spec({
        **params,
        'kty': 'OKP',
        'curve': curve,
        'provider': 'local'
    })
    assert spec.kty == 'OKP'
    assert spec.use == 'enc'
    assert spec.algorithm == "ECDH-ES+A256KW"


@pytest.mark.parametrize("name,cls", [
    ("local", LocalProvider),
    ("random", RandomProvider),
])
def test_lookup_default_providers(name: str, cls: type[core.Provider]):
    provider = core.provider(name)
    assert isinstance(provider, cls)


@pytest.mark.parametrize("name", [
    "local"
])
@pytest.mark.asyncio
async def test_provider_supports_random(name: str):
    provider = core.provider(name)
    assert len(await provider.random(16)) == 16
    assert isinstance(await provider.random_urlsafe(16), str)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import copy
import os
from typing import Any

import pytest
import pytest_asyncio
from jwcrypto.common import json_encode # type: ignore
from jwcrypto.jwe import JWE
from jwcrypto.jwk import JWK
from jwcrypto.jws import JWS

from ckms.core import Keychain
from ckms.utils import b64encode


DATA: bytes = os.urandom(16)


@pytest.fixture
def data() -> bytes:
    return DATA


@pytest.fixture
def jwe():
    payload = b'Hello world!'
    k1 = JWK.generate(kty='oct', size=256)
    k2 = JWK.generate(kty='oct', size=256)
    jwe = JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512"
        })
    )
    jwe.add_recipient(k1, {'foo': 1})
    jwe.add_recipient(k2)
    return jwe


@pytest.fixture
def jwe_single():
    payload = b'Hello world!'
    k1 = JWK.generate(kty='oct', size=256)
    jwe = JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512",
            'foo': 1
        })
    )
    jwe.add_recipient(k1, {'bar': 1})
    return jwe


@pytest.fixture
def jws() -> JWS:
    payload = 'Hello world!'
    k1 = JWK.generate(kty='oct', size=32) # type: ignore
    k2 = JWK.generate(kty='oct', size=32) # type: ignore
    k3 = JWK.generate(kty='oct', size=32) # type: ignore
    k4 = JWK.generate(kty='oct', size=32) # type: ignore
    jws = JWS(payload)
    jws.add_signature(k1, 'HS256', protected={'foo': 1}, header={'bar': 1})
    jws.add_signature(k2, 'HS256', protected={'foo': 2})
    return jws


@pytest.fixture
def jws_single() -> JWS:
    payload = 'Hello world!'
    k1 = JWK.generate(kty='oct', size=32) # type: ignore
    jws = JWS(payload)
    jws.add_signature(k1, 'HS256', protected={'foo': 1, 'alg': 'HS256'}, header={'bar': 1})
    return jws


@pytest.fixture
def multi_jwe(jwe: JWE) -> bytes:
    return b64encode(jwe.serialize(compact=False))


@pytest.fixture
def multi_jws(jws: JWS) -> bytes:
    return b64encode(jws.serialize(compact=False))


@pytest.fixture
def flattened_jwe(jwe_single: JWE) -> bytes:
    return b64encode(jwe_single.serialize(compact=False))


@pytest.fixture
def flattened_jws(jws_single: JWS) -> bytes:
    return b64encode(jws_single.serialize(compact=False))


@pytest.fixture
def compact_jwe(jwe_single: JWE) -> str:
    return jwe_single.serialize(compact=True)


@pytest.fixture
def compact_jws(jws_single: JWS) -> str:
    return jws_single.serialize(compact=True)


@pytest_asyncio.fixture(scope='session')
async def keychain(keys: dict[str, Any]):
    keychain = Keychain()
    keychain.configure(keys=copy.deepcopy(keys))
    await keychain
    return keychain

@pytest_asyncio.fixture(scope='session', autouse=True)
async def setup_keychain(keychain: Keychain):
    await keychain


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`AESOperation`."""
from typing import Any

from cryptography.hazmat.primitives.ciphers import modes
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES

from .ciphertext import CipherText
from .operation import Operation
from .plaintext import PlainText


class AESOperation(Operation):
    content: CipherText | PlainText # type: ignore
    _aad: bytes | None
    _mode: str
    _padding: str | None

    @property
    def aad(self) -> bytes | None:
        return self._aad

    @property
    def mode(self) -> type[modes.CBC] | type[modes.GCM]:
        assert self._mode in {'CBC', 'GCM'}, self._mode
        return getattr(modes, self._mode)

    @property
    def padding(self) -> str | None:
        return self._padding

    def __init__(
        self,
        *,
        mode: str,
        aad: bytes | None = None,
        padding: str | None = None,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        self._mode = mode
        self._aad = aad
        self._padding = padding

    def is_aead(self) -> bool: # pragma: no cover
        return self._mode == 'GCM'

    def get_cipher(
        self,
        iv: bytes | None = None,
        tag: bytes | None = None
    ) -> Cipher[modes.CBC | modes.GCM]:
        if iv is None:
            # This is a decrypt operation and the iv is provided by
            # the input.
            iv = self.get_initialization_vector()
        if iv is None: # pragma: no cover
            raise ValueError("The `iv` parameter can not be None.")
        params: dict[str, Any] = {'initialization_vector': iv}
        if self._mode == 'GCM' and isinstance(self.content, CipherText):
            params['tag'] = self.content.tag
        return Cipher(
            algorithm=self.get_private_key(),
            mode=self.mode(**params)
        )

    def get_ciphertext(self) -> CipherText:
        if not isinstance(self.content, CipherText): # pragma: no cover
            raise TypeError("Input data is not encrypted.")
        return self.content

    def get_initialization_vector(self) -> bytes | None:
        return self.content.iv

    def get_private_key(self) -> AES:
        return self.spec.get_private_key()


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Signature`."""
import inspect
from typing import cast
from typing import Any

import pydantic

from ckms.core.exceptions import MissingProtectedClaim
from ckms.core.exceptions import MissingProtectedHeader
from ckms.types import IKeychain
from ckms.types import Verifier
from ckms.utils import b64decode
from .joseheader_ import JOSEHeader


class Signature(pydantic.BaseModel):
    protected: str
    header: JOSEHeader = pydantic.Field(
        default_factory=JOSEHeader,
        exclude=True
    )
    signature: str

    @property
    def alg(self) -> str | None:
        return self.header.alg

    @property
    def kid(self) -> str | None:
        return self.header.kid

    @pydantic.root_validator(allow_reuse=True, pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        protected = values.get('protected')
        if not protected:
            raise MissingProtectedHeader
        return values

    def json(self, **kwargs: Any) -> dict[str, Any]: # type: ignore
        kwargs.setdefault('exclude', ['header'])
        return super().dict(**kwargs)

    def dict(self, **kwargs: Any) -> dict[str, Any]: # type: ignore
        kwargs.setdefault('exclude', ['header'])
        return super().dict(**kwargs)

    async def verify(
        self,
        keychain: IKeychain,
        payload: str,
        require_kid: bool = True
    ) -> bool:
        """Verifies that this signature was created for the given payload using the
        given `keychain`.
        
        The `require_kid` indicates if the ``kid`` is required to be present on in
        the JOSE header. If `require_kid` is ``True`` and the claim is not present,
        validation fails immediately. Otherwise, all keys in the `keychain` are
        attempted to verify the signature with the specified algorithm.
        """
        if not self.kid and require_kid:
            raise MissingProtectedClaim('kid')
        if not self.alg:
            raise MissingProtectedClaim('alg')
        result = False

        # Construct the signing input and decode the signature
        signing_input = str.encode('.'.join([self.protected, payload]), 'ascii')
        signature = b64decode(self.signature)

        # Use the key identifier to lookup the key. If not present, resort to
        # all registered keys with the given algorithm.
        if self.kid and keychain.has(self.kid):
            spec = cast(Verifier, keychain.get(kid=self.kid))
            if spec.allows_algorithm(self.alg):
                result = spec.verify(
                    signature=signature,
                    message=signing_input
                )
                if inspect.isawaitable(result):
                    result = await result
        elif not require_kid:
            keychain = keychain.filter(
                algorithm=self.alg,
                kid=self.kid,
                use='sig',
                op='verify'
            )
            for spec in keychain.verifiers():
                if not spec.allows_algorithm(self.alg):
                    continue
                result = spec.verify(
                    signature=signature,
                    message=signing_input
                )
                if inspect.isawaitable(result):
                    result = await result
                if result:
                    break
            else:
                result = False
        return result

    def __eq__(self, y: 'Signature') -> bool:
        return self.signature == y.signature



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.type.univ import ObjectIdentifier
from pyasn1_modules import rfc2459


HASHING_ALGORITHM_OIDS: dict[str, ObjectIdentifier] = {
    'sha1'  : ObjectIdentifier((1,3,14,3,2,26)),
    'sha256': ObjectIdentifier((2,16,840,1,101,3,4,2,1,)),
    'sha384': ObjectIdentifier((2,16,840,1,101,3,4,2,2,)),
    'sha512': ObjectIdentifier((2,16,840,1,101,3,4,2,3,))
}


class AlgorithmIdentifier(rfc2459.AlgorithmIdentifier):

    @classmethod
    def new(cls, name: str) -> 'AlgorithmIdentifier':
        """Create a new :class:`AlgorithmIdentifier` by name."""
        instance = cls()
        instance.setComponentByPosition(0, HASHING_ALGORITHM_OIDS[name])
        return instance


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic

from cryptography import x509
from pyasn1_modules import rfc3161

from .pkistatusinfo import PKIStatusInfo
from .timestampreq import TimeStampReq
from .timestamptoken import TimeStampToken
from .tstinfo import TSTInfo


class TimeStampResp(pydantic.BaseModel):
    content: bytes
    request: TimeStampReq
    status: PKIStatusInfo
    token: TimeStampToken | None = pydantic.Field(
        default=None,
        alias='timeStampToken'
    )

    @classmethod
    def parse_asn1(cls, obj: rfc3161.TimeStampResp) -> 'TimeStampResp':
        return cls.parse_obj({
            'status': obj[0],
            'token': obj[1] if obj[1].isValue else None # type: ignore
        })

    @pydantic.validator('token', pre=True)
    def validate_token(
        cls,
        value: rfc3161.TimeStampToken
    ) -> rfc3161.TimeStampToken | None:
        return value if value.isValue else None

    def get_tst_info(self) -> None | TSTInfo:
        """Return a :class:`~ckms.ext.tsa.types.TSTInfo` object
        containing the timestamp metadata, if the signing was
        successful.
        """
        return self.token.info if self.token else None

    def verify(self, certificate: x509.Certificate | None = None) -> None:
        """Verify that a proper response was received from an :rfc:`3161`
        compliant Time Stamping Authority.
        """
        if not self.status.is_success():
            raise ValueError(
                "Timestamp was not granted by the Time Stamping Authority."
            )
        assert self.token is not None
        self.token.verify(
            certificate=certificate,
            request=self.request
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`ClaimSet`."""
import json
from typing import Any
from typing import Generator
from typing import Iterable

import pydantic

from ckms.types import AudienceType
from ckms.utils import b64encode
from ckms.utils import b64decode_json
from .malformed import ClaimTypeError
from .malformed import MalformedPayload
from .malformed import MissingRequiredClaim


class ClaimSet(pydantic.BaseModel):
    __module__: str = 'ckms.types'
    _strict_required: set[str] = set()
    _extra: dict[str, Any] = pydantic.PrivateAttr(default={})

    @property
    def extra(self) -> dict[str,Any]:
        return self._extra

    @classmethod
    def fromjson(cls, payload: bytes | str) -> 'ClaimSet':
        try:
            if isinstance(payload, bytes):
                payload = bytes.decode(payload, 'utf-8')
            return cls.parse_obj(
                json.loads(payload)
            )
        except (ValueError, TypeError):
            raise MalformedPayload

    @classmethod
    def frompayload(cls, payload: bytes | str) -> 'ClaimSet':
        return cls.parse_obj(
            b64decode_json(
                buf=payload,
                on_failure=MalformedPayload()
            )
        )

    @classmethod
    def strict(cls, **claims: Any) -> 'ClaimSet':
        raise NotImplementedError

    def __init__(self, **claims: Any):
        try:
            extra = claims.pop('extra', {})
            super().__init__(**claims)
            for name, value in claims.items():
                if name in self.__fields__:
                    continue
                extra[name] = value
            self._extra = extra
        except pydantic.ValidationError as exception:
            for error in exception.errors():
                error_type = error.get('type') or ''
                if str.startswith(error_type, 'type_error'):
                    claim = error['loc'][0]
                    raise ClaimTypeError(str(claim)) from exception
            else:
                raise

    def clone(self, values: dict[str, Any]) -> 'ClaimSet':
        """Clone the claims set with the updated `values`."""
        return type(self).parse_obj({
            **self.dict(),
            **values
        })

    def dict(self, **kwargs: Any) -> dict[str, Any]: # type: ignore
        kwargs.setdefault('exclude_defaults', True)
        return super().dict(**kwargs)

    def json(self, **kwargs: Any) -> str: # type: ignore
        kwargs.setdefault('exclude_defaults', True)
        return super().json(**kwargs)

    def verify(
        self,
        *,
        required: set[str] | None = None,
        strict: bool = False,
    ) -> None:
        """Verifies the claims according to the specification laid out
        in the JOSE RFCs.
        """
        required = required or set()
        if strict:
            required.update(self._strict_required)
        for claim in required:
            if getattr(self, claim, self.extra.get(claim)) is not None:
                continue
            raise MissingRequiredClaim(claim)

    def _iter(self, **kwargs: Any) -> Generator[tuple[str, Any], None, None]:
        for name, value in super()._iter(**kwargs):
            if name != 'extra':
                yield name, value
                continue
        for extra_claim, extra_value in self.extra.items():
            yield extra_claim, extra_value

    def __bytes__(self) -> bytes:
        return b64encode(self.json())

    def __xor__(self, value: set[str] | Iterable[Any]) -> set[str]:
        if not isinstance(value, set):
            value = set(value)
        return set(self.__fields__.keys()) ^ value

    def __rxor__(self, value: set[str]) -> set[str]:
        return self.__xor__(value)

    class Config:
        json_encoders: dict[type[Any], Any] = {
            AudienceType: lambda aud: list(sorted(aud)) # type: ignore
        }


# pylint: skip-file
import pytest

import ckms


PROJECT = "unimatrixdev"

LOCATION = "europe-west4"

KEYRING = "local"


SIGNING_ALGORITHMS = [
    ('rsa_pkcs_2048', 'RS256'),
    ('rsa_pkcs_3072', 'RS256'),
    ('rsa_pkcs_4096', 'RS256'),
    ('rsa_pkcs_4096_sha512', 'RS512'),
    ('rsa_pss_2048', 'PS256'),
    ('rsa_pss_3072', 'PS256'),
    ('rsa_pss_4096', 'PS256'),
    ('rsa_pss_4096_sha512', 'PS512'),
    ('hmacsha256', 'HS256'),
    ('p256', 'ES256'),
    ('p384', 'ES384'),
    ('p256k', 'ES256K'),
]


@pytest.fixture(scope='session')
def keychain():
    return ckms.Keychain()


@pytest.fixture(scope='session')
def keyconfig():
    return {
        'p256': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "ec_sign_p256_sha256"
        },
        'p384': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "ec_sign_p384_sha384"
        },
        'p256k': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "ec_sign_secp256k1_sha256"
        },
        'hmacsha256': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "hmac_sha256"
        },
        'rsa_pkcs_2048': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pkcs1_2048_sha256"
        },
        'rsa_pkcs_3072': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pkcs1_3072_sha256"
        },
        'rsa_pkcs_4096': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pkcs1_4096_sha256"
        },
        'rsa_pkcs_4096_sha512': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pkcs1_4096_sha512"
        },
        'rsa_pss_2048': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pss_2048_sha256"
        },
        'rsa_pss_3072': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pss_3072_sha256"
        },
        'rsa_pss_4096': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pss_4096_sha256"
        },
        'rsa_pss_4096_sha512': {
            'provider': "google",
            'project': PROJECT,
            'location': LOCATION,
            'keyring': KEYRING,
            'key': "rsa_sign_pss_4096_sha512"
        },
    }



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from httpx import AsyncClient

from ckms.ext.tsa import Timestamper


@pytest.fixture
def timestamper() -> Timestamper:
    return Timestamper.fromlist([
        {'url': "http://tss.accv.es:8318/tsa"},
        {'url': "http://timestamp.apple.com/ts01"},
        {'url': "http://timestamp.digicert.com"},
        {'url': "http://aatl-timestamp.globalsign.com/tsa/aohfewat2389535fnasgnlg5m23"},
        {'url': "http://tsa.izenpe.com"},
        {'url': "http://kstamp.keynectis.com/KSign/"},
        {'url': "http://tsa.quovadisglobal.com/TSS/HttpTspServer"},
    ])


@pytest.mark.asyncio
async def test_timestamp_with_default_signatures(
    data: bytes,
    client: AsyncClient,
    timestamper: Timestamper
):
    responses = await timestamper.timestamp(client, data)
    assert len(responses) == 1


@pytest.mark.asyncio
async def test_timestamp_with_min_signatures(
    data: bytes,
    client: AsyncClient,
    timestamper: Timestamper
):
    responses = await timestamper.timestamp(client, data, min_signatures=3)
    assert len(responses) == 3


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__: list[str] = [
    'ID_CE_BASICCONSTRAINTS',
    'ID_CE_EXTKEYUSAGE',
    'ID_CE_KEYUSAGE',
    'ID_CE_SUBJECTALTNAME',
    'ID_ECDH',
    'ID_EC_PUBLICKEY',
    'ID_EXTENSION_REQUEST',
]


ID_CE_BASICCONSTRAINTS: str = '2.5.29.19'

ID_CE_EXTKEYUSAGE: str = '2.5.29.37'

ID_CE_KEYUSAGE: str = '2.5.29.15'

ID_CE_SUBJECTALTNAME: str = '2.5.29.17'

ID_EC_PUBLICKEY: str = '1.2.840.10045.2.1'

ID_ECDH: str = '1.3.132.1.12'

ID_EXTENSION_REQUEST: str = '1.2.840.113549.1.9.14'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import pytest_asyncio

from ckms.core import parse_spec
from ckms.core.models import KeySpecification
from ckms.types import JSONWebKey


@pytest_asyncio.fixture(scope='session') # type: ignore
async def key() -> KeySpecification:
    return await parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'ec_sign_p256_sha256',
        'version': 5,
        'certificate': {
            'uri': 'https://certs.unimatrixone.io/DevelopmentTestCertificateG5.crt'
        }
    })


@pytest.fixture(scope='session')
def jwk(key: KeySpecification) -> JSONWebKey:
    return JSONWebKey.parse_obj(key.as_public().as_jwk(private=False))


def test_key_has_x5c(jwk: JSONWebKey):
    assert jwk.x5c is not None


def test_key_has_x5t(jwk: JSONWebKey):
    assert jwk.x5t is not None


def test_key_has_x5t_sha256(jwk: JSONWebKey):
    assert jwk.x5t_sha256 is not None


def test_key_has_x5u(jwk: JSONWebKey):
    assert jwk.x5u is not None


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .attribute import Attribute
from .attributelist import AttributeList
from .certificateextension import CertificateExtension
from .basicconstraints import BasicConstraints
from .certificateextension import KeyUsage
from .certificationrequest import CertificationRequest
from .certificateextensionsequence import CertificateExtensionSequence


__all__: list[str] = [
    'Attribute',
    'AttributeList',
    'BasicConstraints',
    'CertificateExtension',
    'CertificateExtensionSequence',
    'CertificationRequest',
    'KeyUsage'
]


# pylint: skip-file
import asyncio
import copy
import os
from typing import Any

import pytest
import pytest_asyncio

import ckms.core
from ckms.core import Keychain


@pytest.fixture(scope="session")
def event_loop():
    return asyncio.new_event_loop()


@pytest.fixture
def data():
    return os.urandom(16)


@pytest.fixture(scope='session')
def keyconfig() -> dict[str, Any]:
    return {}


@pytest.fixture(scope='session')
def asymmetric_sig_keys() -> dict[str, Any]:
    return {
        'EdDSA': {
            'provider': 'local',
            'kty': 'OKP',
            'use': 'sig'
        },
        'ES256': {
            'provider': 'local',
            'kty': 'EC',
            'algorithm': 'ES256'
        },
        'ES256K': {
            'provider': 'local',
            'kty': 'EC',
            'algorithm': 'ES256K'
        },
        'ES384': {
            'provider': 'local',
            'kty': 'EC',
            'algorithm': 'ES384'
        },
        'ES512': {
            'provider': 'local',
            'kty': 'EC',
            'algorithm': 'ES512'
        },
        'RS256': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RS256',
            'key': {'path': 'pki/RS256.key'}
        },
        'RS384': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RS384',
            'key': {'path': 'pki/RS384.key'}
        },
        'RS512': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RS512',
            'key': {'path': 'pki/RS512.key'}
        },
        'PS256': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'PS256',
            'key': {'path': 'pki/PS256.key'}
        },
        'PS384': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'PS384',
            'key': {'path': 'pki/PS384.key'}
        },
        'PS512': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'PS512',
            'key': {'path': 'pki/PS512.key'}
        },
    }


@pytest.fixture(scope='session')
def asymmetric_enc_keys() -> dict[str, Any]:
    return {
        'ECDH-ES': {
            'provider': 'local',
            'kty': 'OKP',
            'algorithm': 'ECDH-ES',
            'use': 'enc'
        },
        'ECDH-ES+A128KW': {
            'provider': 'local',
            'kty': 'OKP',
            'algorithm': 'ECDH-ES+A128KW',
            'use': 'enc'
        },
        'ECDH-ES+A192KW': {
            'provider': 'local',
            'kty': 'OKP',
            'algorithm': 'ECDH-ES+A192KW',
            'use': 'enc'
        },
        'ECDH-ES+A256KW': {
            'provider': 'local',
            'kty': 'OKP',
            'algorithm': 'ECDH-ES+A256KW',
            'use': 'enc'
        },
        'RSA-OAEP': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA-OAEP',
            'key': {'path': 'pki/RSA-OAEP.key'}
        },
        'RSA-OAEP-256': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA-OAEP-256',
            'key': {'path': 'pki/RSA-OAEP-256.key'}
        },
        'RSA-OAEP-384': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA-OAEP-384',
            'key': {'path': 'pki/RSA-OAEP-384.key'}
        },
        'RSA-OAEP-512': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA-OAEP-512',
            'key': {'path': 'pki/RSA-OAEP-512.key'}
        },
        'RSA1_5': {
            'provider': 'local',
            'kty': 'RSA',
            'algorithm': 'RSA1_5',
            'key': {'path': 'pki/RSA1_5.key'}
        },
    }


@pytest.fixture(scope='session')
def symmetric_sig_keys() -> dict[str, Any]:
    return {
        'HS256': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS256',
            'key': {'length': 32}
        },
        'HS384': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS384',
            'key': {'length': 32}
        },
        'HS512': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'HS512',
            'key': {'length': 32}
        },
    }


@pytest.fixture(scope='session')
def symmetric_enc_keys() -> dict[str, Any]:
    return {
        'A128GCM': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A128GCM',
            'key': {'length': 16}
        },
        'A192GCM': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A192GCM',
            'key': {'length': 24}
        },
        'A256GCM': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A256GCM',
            'key': {'length': 32}
        },
        'A128GCMKW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A128GCMKW',
            'key': {'length': 16}
        },
        'A192GCMKW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A192GCMKW',
            'key': {'length': 24}
        },
        'A256GCMKW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A256GCMKW',
            'key': {'length': 32}
        },
        'A128KW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A128KW',
            'key': {'length': 16}
        },
        'A192KW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A192KW',
            'key': {'length': 24}
        },
        'A256KW': {
            'provider': 'local',
            'kty': 'oct',
            'algorithm': 'A256KW',
            'key': {'length': 32}
        },
    }


@pytest.fixture(scope='session')
def asymmetric_keys(
    asymmetric_sig_keys: dict[str, Any],
    asymmetric_enc_keys: dict[str, Any],
) -> dict[str, Any]:
    return {
        **asymmetric_sig_keys,
        **asymmetric_enc_keys
    }


@pytest.fixture(scope='session')
def symmetric_keys(
    symmetric_sig_keys: dict[str, Any],
    symmetric_enc_keys: dict[str, Any],
) -> dict[str, Any]:
    return {
        **symmetric_sig_keys,
        **symmetric_enc_keys
    }


@pytest.fixture(scope='session')
def keys(
    asymmetric_keys: dict[str, Any],
    symmetric_keys: dict[str, Any],
) -> dict[str, Any]:
    return {
        **asymmetric_keys,
        **symmetric_keys
    }


@pytest_asyncio.fixture(scope='session') # type: ignore
async def keychain(keys: dict[str, Any]):
    keychain = Keychain()
    keychain.configure(keys=copy.deepcopy(keys))
    await keychain
    return keychain


@pytest_asyncio.fixture(scope='session', autouse=True) # type: ignore
async def setup_keychain(keychain: Keychain):
    await keychain


@pytest_asyncio.fixture(scope='session') # type: ignore
async def hs256():
    key = ckms.core.parse_spec({
        'provider': 'local',
        'kty': 'oct',
        'algorithm': 'HS256',
        'key': {'length': 32}
    })
    await key
    return key


@pytest_asyncio.fixture(scope='session') # type: ignore
async def hs384():
    key = ckms.core.parse_spec({
        'provider': 'local',
        'kty': 'oct',
        'algorithm': 'HS384',
        'key': {'length': 32}
    })
    await key
    return key


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic
from pyasn1.type import univ
from pyasn1_modules import rfc5280

from ckms.ext.x509.types import AlgorithmIdentifier


class MessageImprint(pydantic.BaseModel):
    algorithm: AlgorithmIdentifier = pydantic.Field(..., alias='hashAlgorithm')
    message: bytes = pydantic.Field(..., alias='hashedMessage')

    @pydantic.validator('algorithm', pre=True)
    def validate_algorithm(
        cls,
        value: rfc5280.AlgorithmIdentifier | str | AlgorithmIdentifier
    ) -> str | AlgorithmIdentifier:
        if isinstance(value, rfc5280.AlgorithmIdentifier):
            value = str(value['algorithm']) # type: ignore
        return value

    @pydantic.validator('message', pre=True)
    def validate_message(cls, value: bytes | univ.OctetString) -> bytes:
        return bytes(value)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib

import pytest

from ckms.core import fetch
from ckms.core import RemoteBlob


@pytest.fixture
def blob():
    return RemoteBlob(
        provider='local',
        path=pathlib.Path(__file__).parent.joinpath('test.txt')
    )


@pytest.mark.asyncio
async def test_get_content(blob: RemoteBlob):
    assert await fetch(blob) == b'Hello world!'



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Signature`."""
import json

from typing import Any

from ckms.types import Data
from ckms.types import Signer
from ckms.utils import b64encode


class Signature:
    __module__: str = 'ckms.jose'
    header: dict[str, Any]
    payload: Data
    protected: dict[str, Any]
    protected_b64: bytes | None = None
    signer: Signer
    signature: bytes | None = None

    def __init__(
        self,
        *,
        protected: dict[str, Any],
        signer: Signer,
        payload: Data,
        header: dict[str, Any] | None = None
    ):
        self.header = header or {}
        self.payload = payload
        self.protected = protected
        self.signer = signer

        protected.update({
            'alg': signer.algorithm
        })

    async def sign(
        self,
        payload: bytes,
        protected: dict[str, Any] | None = None,
        compact: bool = False
    ) -> 'Signature':
        """Create a signature of the protected header and the payload using the
        signing key supplied at initialization.
        """
        protected = {**(protected or {}), **self.protected}
        if compact and self.header:
            protected = {**self.header, **protected}
        if self.signature is None:
            self.protected_b64 = b64encode(
                json.dumps(protected, sort_keys=True).encode('utf-8')
            )
            message = b'.'.join([self.protected_b64, payload])
            self.signature = b64encode(await self.signer.sign(message))
            assert bool(self)
        return self

    def dict(self, compact: bool) -> dict[str, Any]:
        """Dump the JSON Web Signature (JWS) using the JWS JSON Serialization
        schema.
        """
        assert self.protected_b64 is not None
        if self.signature is None:
            raise ValueError("Invoke Signature.sign() before serializing.")
        jws: dict[str, Any] = {
            'protected': bytes.decode(self.protected_b64, 'ascii'),
            'signature': bytes.decode(self.signature, 'ascii')
        }
        if not compact and bool(self.header):
            jws['header'] = self.header
        return jws

    def __bool__(self) -> bool:
        return bool(self.signature)

    def __repr__(self) -> str:
        return f'<Signarure: signed={bool(self)}>'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import pathlib
from typing import Any

import pytest

from ckms.types import Data
from ckms.utils import b64decode


FILEPATH: pathlib.Path = pathlib.Path(__file__).parent.joinpath('test.txt')

DATA: bytes = open(FILEPATH, 'rb').read()

INPUTS: list[dict[str, Any]] = [
    {'buf': b'Hello world!'},
    {'src': FILEPATH},
    {'src': str(FILEPATH)}
]

ENCODINGS: list[str] = [
    'base64'
]

HASHING_ALGORITHMS = [
    ('sha256', hashlib.sha256),
    ('sha384', hashlib.sha384),
    ('sha512', hashlib.sha512),
    ('sha3_256', hashlib.sha3_256),
    ('sha3_384', hashlib.sha3_384),
    ('sha3_512', hashlib.sha3_512),
]


class BaseDataTestCase:
    impl: type[Data]

    def get_content(self, content: bytes) -> bytes:
        return content

    def get_digest(self, factory: Any, data: Data) -> bytes:
        return factory(bytes(data)).digest() # type: ignore

    @pytest.mark.parametrize("params", INPUTS)
    @pytest.mark.asyncio
    async def test_iter(self, params: dict[str, Any]):
        data = self.impl(**params)
        async for chunk in data.chunks(length=1):
            assert len(chunk) == 1

    @pytest.mark.parametrize("params", INPUTS)
    @pytest.mark.asyncio
    async def test_iter_content(self, params: dict[str, Any]):
        data = self.impl(**params)
        content = b''
        async for chunk in data.chunks(length=1):
            content += chunk
        assert content == self.get_content(b'Hello world!')

    @pytest.mark.parametrize("hasher", HASHING_ALGORITHMS)
    @pytest.mark.parametrize("params",INPUTS)
    @pytest.mark.asyncio
    async def test_digest(
        self,
        hasher: Any,
        params: dict[str, Any]
    ):
        algorithm, factory = hasher
        data = self.impl(**params)
        digest = await data.digest(algorithm=algorithm)
        hasher = factory(bytes(data))
        assert len(digest) == len(hasher.digest())
        assert digest == self.get_digest(factory, data)

    def test_src_or_buf_is_required(self):
        with pytest.raises(TypeError):
            self.impl()

    def test_src_and_buf_is_exception(self):
        with pytest.raises(TypeError):
            self.impl(src='foo.txt', buf=b'')

    @pytest.mark.parametrize("hasher", HASHING_ALGORITHMS)
    @pytest.mark.parametrize("params",INPUTS)
    @pytest.mark.asyncio
    async def test_algorithm_from_constructor(
        self,
        hasher: Any,
        params: dict[str, Any]
    ):
        algorithm, factory = hasher
        data = self.impl(digest=algorithm, **params)
        assert await data.digest() == self.get_digest(factory, data)

    @pytest.mark.asyncio
    async def test_algorithm_is_required_if_not_in_constructor(self):
        data = self.impl(buf=DATA)
        with pytest.raises(TypeError):
            await data.digest()

    @pytest.mark.parametrize("algorithm", [x[0] for x in HASHING_ALGORITHMS])
    @pytest.mark.parametrize("encoding", ENCODINGS)
    @pytest.mark.parametrize("verify", [b64decode])
    @pytest.mark.parametrize("params", INPUTS)
    @pytest.mark.asyncio
    async def test_encode_digest(
        self,
        algorithm: str,
        encoding: str,
        verify: Any,
        params: dict[str, Any]
    ):
        data = self.impl(**params)
        digest = await data.digest(
            algorithm=algorithm,
            encoding=encoding
        )
        assert verify(digest) == await data.digest(algorithm=algorithm)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .contentencryptionkey import ContentEncryptionKey
from .edwardscurvekeyspecification import EdwardsCurveKeySpecification
from .ellipticcurvekeyspecification import EllipticCurveKeySpecification
from .key import Key
from .localkey import LocalKey
from .localkeyspecification import LocalKeySpecification
from .hmac import HMAC
from .parameterlesskey import ParameterLessKey
from .rsakeyspecification import RSAKeySpecification
from .symmetrickeyspecification import SymmetricKeySpecification
from .transientkey import TransientKey


__all__: list[str] = [
    'ContentEncryptionKey',
    'EdwardsCurveKeySpecification',
    'EllipticCurveKeySpecification',
    'HMAC',
    'Key',
    'LocalKey',
    'LocalKeySpecification',
    'ParameterLessKey',
    'RSAKeySpecification',
    'SymmetricKeySpecification',
    'TransientKey',
]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unimatrix.exceptions import CanonicalException

from .keyoperationtype import KeyOperationType


class ForbiddenOperation(CanonicalException):
    __module__: str = 'ckms.types'
    http_status_code: int = 403
    code: str = 'CRYPTOGRAPHIC_OPERATION_FORBIDDEN'
    message: str = "Unable to perform the requested cryptographic operation."

    def __init__(self, op: KeyOperationType):
        super().__init__( # type: ignore
            detail=(
                "The request was rejected because the cryptographic key "
                f"is not allowed to perform the '{op.value}' operation."
            )
        )
        self.op = op


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from pyasn1.codec.der import decoder
from pyasn1.type.char import PrintableString
from pyasn1.type.char import UTF8String
from pyasn1_modules import rfc2459


class RelativeDistinguishedName(pydantic.BaseModel):
    common_name: str | None = pydantic.Field(
        default=None,
        alias='2.5.4.3'
    )

    country: str | None = pydantic.Field(
        default=None,
        alias='2.5.4.6'
    )

    organization_name: str | None = pydantic.Field(
        default=None,
        alias='2.5.4.10'
    )

    @classmethod
    def from_attrs(cls, rdn: rfc2459.RelativeDistinguishedName) -> 'RelativeDistinguishedName':
        attrs: dict[str, str] = {}
        for attr in rdn: # type: ignore
            attrs[str(attr['type'])] = attr['value'] # type: ignore
        return cls.parse_obj(attrs)

    @pydantic.root_validator(pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        for oid, value in list(values.items()):
            if not isinstance(value, rfc2459.AttributeValue):
                continue
            obj, _ = decoder.decode(value) # type: ignore
            if _:
                raise ValueError("Invalid DER encoding.")
            if isinstance(obj, (PrintableString, UTF8String)):
                values[oid] = str(obj)
            else:
                raise NotImplementedError(type(obj)) # type: ignore

        return values


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`ServerMetadata`."""
import inspect
import logging
import typing
import urllib.parse
from typing import cast
from typing import Any
from typing import Callable

import pydantic

from .ihttpclient import IHTTPClient
from .jsonwebkeyset import JSONWebKeySet
from .trustissues import AuthorizationServerMisbehaves
from .trustissues import AuthorizationServerNotDiscoverable
from .trustissues import UnresolvableIssuer


AUTH_METHODS = [
    "none",
    "client_secret_post",
    "client_secret_basic",
    "client_secret_jwt",
    "private_key_jwt",
    "tls_client_auth",
    "self_signed_tls_client_auth",
]


CONTENT_ENCRYPTION_ALGORITHMS = [
    "A128CBC-HS256",
    "A192CBC-HS384",
    "A256CBC-HS512",
    "A128GCM",
    "A192GCM",
    "A256GCM",
]

ENCRYPTION_ALGORITHMS = [
    "RSA1_5",
    "RSA-OAEP",
    "RSA-OAEP-256",
    "A128KW",
    "A192KW",
    "A256KW",
    "dir",
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "PBES2-HS256+A128KW",
    "PBES2-HS384+A192KW",
    "PBES2-HS512+A256KW",
]

SIGNATURE_ALGORITHMS = [
    "HS256",
    "HS384",
    "HS512",
    "RS256",
    "RS384",
    "RS512",
    "PS256",
    "PS384",
    "PS512",
    "ES256",
    "ES384",
    "ES512",
    "ES256K",
    "EdDSA"
]


class ServerMetadata(pydantic.BaseModel):
    issuer: str = pydantic.Field(
        default=None,
        title="Issuer",
        description="Authorization server's issuer identifier URL.",
        example="https://webid.unimatrixapis.com"
    )

    authorization_endpoint: str = pydantic.Field(
        default=None,
        title="Authorization endpoint",
        description="URL of the authorization server's authorization endpoint.",
        example="https://webid.unimatrixapis.com/authorize"
    )

    token_endpoint: str = pydantic.Field(
        default=None,
        title="Token endpoint",
        description="URL of the authorization server's token endpoint.",
        example="https://oauth2.unimatrixapis.com/token"
    )

    jwks_uri: typing.Optional[str] = pydantic.Field(
        default=None,
        title="JSON Web Key Set (JWKS) URI",
        description="URL of the authorization server's JWK Set document.",
        example="https://oauth2.unimatrixapis.com/jwks"
    )

    registration_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Registration endpoint",
        description=(
            "URL of the authorization server's **OAuth 2.0 Dynamic Client "
            "Registration Endpoint**."
        ),
        example="https://oauth2.unimatrixapis.com/register"
    )

    scopes_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported scopes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `scope` values that "
            "this authorization server supports."
        ),
        example=["foo", "bar", "baz"]
    )

    response_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported response types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_type` "
            "values that this authorization server supports."
        ),
        example=[
            "code",
            "code id_token",
            "code id_token token",
            "code token",
            "id_token",
            "id_token token",
            "none",
            "token"
        ]
    )

    response_modes_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported response modes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_mode` "
            "values that this authorization server supports."
        ),
        example=[
            "fragment",
            "query",
        ]
    )

    grant_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported grant types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `grant_types` "
            "values that this authorization server supports."
        ),
        example=[
            "authorization_code",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:device_code",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "urn:ietf:params:oauth:grant-type:saml2-bearer",
            "urn:ietf:params:oauth:grant-type:token-exchange",
        ]
    )

    token_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported client authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this token endpoint."
        ),
        example=AUTH_METHODS
    )

    token_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the token endpoint for the signature on the JWT "
            "used to authenticate the client at the token endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    service_documentation: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Documentation",
        description=(
            "URL of a page containing human-readable information that "
            "developers might want or need to know when using the "
            "authorization server."
        ),
        example="https://oauth2.unimatrixapis.com/docs"
    )

    ui_locales_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported locals for UI",
        description=(
            "Languages and scripts supported for the user interface, "
            "represented as a JSON array of language tag values from BCP 47."
        ),
        example=["nl-NL, be"]
    )

    op_policy_uri: typing.Optional[str]= pydantic.Field(
        default=None,
        title="Data policy URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "requirements on how the client can use the data provided by the "
            "authorization server."
        ),
        example="https://oauth2.unimatrixapis.com/privacy"
    )

    op_tos_uri: typing.Optional[str]= pydantic.Field(
        default=None,
        title="Terms of service URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "terms of service."
        ),
        example="https://oauth2.unimatrixapis.com/tos"
    )

    revocation_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Revocation endpoint",
        description="URL of the authorization server's revocation endpoint.",
        example="https://oauth2.unimatrixapis.com/revoke"
    )

    revocation_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this revocation endpoint."
        ),
        example=AUTH_METHODS
    )

    revocation_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the revocation endpoint for the signature on the JWT "
            "used to authenticate the client at the revocation endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    introspection_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Introspection endpoint",
        description="URL of the authorization server's introspection endpoint.",
        example="https://oauth2.unimatrixapis.com/introspect"
    )

    introspection_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this introspection endpoint."
        ),
        example=AUTH_METHODS
    )

    introspection_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the introspection endpoint for the signature on the JWT "
            "used to authenticate the client at the introspection endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    signed_metadata: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "Signed JWT containing metadata values about the authorization "
            "server as claims."
        )
    )


    device_authorization_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Device authorization endpoint",
        description=(
            "URL of the authorization server's device authorization endpoint."
        ),
        example="https://oauth2.unimatrixapis.com/device"
    )

    tls_client_certificate_bound_access_tokens: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Supports mTLS certificate-bound access tokens",
        description=(
            "Indicates authorization server support for mutual-TLS client "
            "certificate-bound access tokens."
        ),
        example=True
    )

    mtls_endpoint_aliases: typing.Optional[typing.Dict[str, str]] = pydantic.Field(
        default={},
        title="Alternative mTLS endpoints",
        description=(
            "JSON object containing alternative authorization server "
            "endpoints, which a client intending to do mutual TLS will "
            "use in preference to the conventional endpoints."
        ),
        example={
            "authorize": "https://webid.unimatrixapis.com/authorize"
        }
    )

    nfv_token_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as NFV Token."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    nfv_token_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    nfv_token_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    userinfo_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="UserInfo endpoint",
        description="URL of the authorization servers' UserInfo Endpoint.",
        example="https://webid.unimatrixapis.com/userinfo"
    )

    acr_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported ACR",
        description=(
            "JSON array containing a list of the Authentication Context Class "
            "References that this authorization server supports."
        ),
        example=["a.b", "a"]
    )

    subject_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported subject types",
        description=(
            "JSON array containing a list of the Subject Identifier types that "
            "this authorization server supports"
        ),
        example=["pairwise", "public"]
    )

    id_token_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as ID Token."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    id_token_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    id_token_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    userinfo_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as UserInfo Endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    userinfo_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    userinfo_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    request_object_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as Request Object."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    request_object_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    request_object_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    display_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported display modes",
        description=(
            "JSON array containing a list of the `display` parameter values "
            "that the OpenID Provider supports."
        ),
        example=["page", "popup", "touch", "wap"]
    )

    claim_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claims Types "
            "that the OpenID Provider supports."
        )
    )

    claims_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claim Names of the Claims "
            "that the OpenID Provider MAY be able to supply values for."
        )
    )

    claims_locales_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported claim locales",
        description=(
            "Languages and scripts supported for values in Claims being "
            "returned, represented as a JSON array of BCP 47."
        ),
        example=["nl-NL", "be"]
    )

    claims_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `claims` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`claims` parameter."
        ),
        example=True
    )

    request_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `request` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request` parameter."
        ),
        example=True
    )

    request_uri_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `request_uri` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request_uri` parameter."
        ),
        example=True
    )

    require_request_uri_registration: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Boolean value specifying whether the OP requires any `request_uri` "
            "values used to be pre-registered."
        ),
        example=True
    )

    require_signed_request_object: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Indicates where authorization request needs to be protected as "
            "**Request Object** and provided through either `request` or "
            "`request_uri` parameter."
        ),
        example=True
    )

    pushed_authorization_request_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Pushed Authorization Request (PAR) endpoint",
        description=(
            "URL of the authorization server's pushed authorization request "
            "endpoint."
        ),
        example="https://oauth2.unimatrixapis.com/par"
    )

    require_pushed_authorization_requests: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Requires PAR?",
        description=(
            "Indicates whether the authorization server accepts authorization "
            "requests only via PAR."
        ),
        example=False
    )

    introspection_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response signing."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    introspection_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content key "
            "encryption (`alg` value)."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    introspection_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content "
            "encryption (`enc` value)."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    authorization_response_iss_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Supports `iss` parameter in authorization response?",
        description=(
            "Boolean value indicating whether the authorization server "
            "provides the `iss` parameter in the authorization response."
        ),
        example=True
    )

    # TODO: These are non-standard properties and we must submit an RFC
    # to register these at IANA.
    token_endpoint_auth_encryption_alg_values_supported: typing.Optional[typing.List[str]] = []
    token_endpoint_auth_encryption_enc_values_supported: typing.Optional[typing.List[str]] = []
    assertion_signing_alg_values_supported: typing.Optional[typing.List[str]] = []
    assertion_encryption_alg_values_supported: typing.Optional[typing.List[str]] = []
    assertion_encryption_enc_values_supported: typing.Optional[typing.List[str]] = []
    required_encrypted_token_endpoint_auth: typing.Optional[bool] = False
    require_encrypted_assertion: typing.Optional[bool] = False

    @classmethod
    async def discover(
        cls,
        client: Any,
        issuer: str,
        path: str | None = None,
        on_failure: type[Exception] | Exception | Callable[[Exception], Exception] | None = None,
        extra_locations: list[str] | None = None,
        logger: logging.Logger | str = 'ckms.auth',
        metadata_url: str | None = None
    ) -> 'ServerMetadata':
        """Discover the metadata of an authorization server using the well-known
        endpoints.
        """
        client = cast(IHTTPClient, client)
        if not isinstance(logger, logging.Logger):
            logger = logging.getLogger(logger)
        path = f'{str.strip(path or "", "/")}/' if path is not None else ''
        extra_locations = extra_locations or []
        p = urllib.parse.urlparse(metadata_url or issuer)
        if not p.netloc or not p.scheme:
            # If there is no scheme or netloc, then the issuer is not a
            # URL, making us unable to construct the well-known metadata
            # URLs.
            raise UnresolvableIssuer
        if len(p.netloc) > 253:
            # If the domain name is longer than 253 characters, it is invalid
            # and should be considered an injection attempt.
            logger.critical("Issuer domain name exceeds maximum length")
            raise UnresolvableIssuer
        locations: list[str] = [metadata_url] if metadata_url is not None else [
            *[f'{issuer}/{str.lstrip(x, "/")}' for x in extra_locations],
            f'{issuer}/{path}.well-known/oauth-authorization-server',
            f'{issuer}/{path}.well-known/openid-configuration',
        ]
        try:
            for url in locations:
                response = client.get(
                    url=url
                )
                if inspect.isawaitable(response):
                    response = await response
                if response.status_code != 200:
                    continue
                payload = response.json()
                if inspect.isawaitable(payload):
                    payload = await payload
                break
            else:
                payload = None
        except Exception as exception:
            logger.critical(
                "Failed to retrieve authorization server metadata (%s)",
                repr(exception)
            )
            if on_failure is None:
                raise
            if callable(on_failure):
                on_failure = on_failure(exception)
            raise on_failure from exception

        if payload is None:
            raise AuthorizationServerNotDiscoverable

        try:
            return cls.parse_obj(payload)
        except pydantic.ValidationError:
            raise AuthorizationServerMisbehaves(
                hint=(
                    "The metadata returned from the well-known configuration "
                    "endpoint was of invalid format."
                )
            )

    async def get_jwks(
        self,
        client: Any
    ) -> JSONWebKeySet | None:
        """Return a :class:`~ckms.types.JSONWebKeySet` instance retrieved
        from the JWKS URI specified by the authorization server, or
        ``None``.
        """
        client = cast(IHTTPClient, client)
        if self.jwks_uri is not None:
            response = client.get(self.jwks_uri)
            if inspect.isawaitable(response):
                response = await response
            jwks = response.json()
            if inspect.isawaitable(jwks):
                jwks = await jwks
            return JSONWebKeySet.parse_obj(jwks)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`SigningOperation`."""
from typing import Any

from .operation import Operation


class SigningOperation(Operation):
    """The base class for all signing operations."""
    __module__: str = 'ckms.core.types'
    _signature: bytes | None

    @property
    def signature(self) -> bytes | None:
        return self._signature

    def __init__(self, *, signature: bytes | None = None, **kwargs: Any):
        super().__init__(**kwargs)
        self._signature = signature


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from pyasn1.type import char
from pyasn1_modules import rfc5280


class Name(pydantic.BaseModel):
    kind: str
    value: Any

    @classmethod
    def parse_value(cls, value: str | char.IA5String) -> Any:
        if isinstance(value, char.IA5String):
            value = str(value)
        return value

    @pydantic.validator('value', pre=True)
    def preprocess_value(cls, value: Any):
        return cls.parse_value(value)

    def encode_value(self) -> Any:
        value = self.value
        if hasattr(self.value, 'asn1'):
            value = value.asn1()
        return value

    def asn1(self) -> rfc5280.GeneralName:
        obj = rfc5280.GeneralName()
        obj[self.kind] = self.encode_value()
        return obj


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`AudienceType`"""
from typing import Any
from typing import Generator
from typing import TypeAlias

from .malformed import ClaimTypeError


VALID_TYPES: TypeAlias = str | set[str] | None


class AudienceType(set[str]):
    __module__: str = "ckms.types"

    @classmethod
    def __get_validators__(
        cls
    ) -> Generator[Any, Any, Any]:
        yield cls.validate

    @classmethod
    def validate(cls,
        value: VALID_TYPES
    ) -> 'AudienceType':
        if isinstance(value, str):
            value = {value}
        if not all([isinstance(x, str) for x in value]): # type: ignore
            raise ClaimTypeError(
                claim='aud',
                hint=(
                    "The 'aud' claim is either a string or an array of strings."
                )
            )
        return cls(value or {})

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None: # pragma: no cover
        field_schema.update(
            type="array",
            examples=[
                "https://oauth2.unimatrixapis.com",
                ["https://oauth2.unimatrixapis.com"]
            ]
        )

    def __repr__(self) -> str:
        return repr(set(self))


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONB64`"""
import json
from typing import Any
from typing import Callable
from typing import Generator

from ckms.utils import b64decode


class JSONB64(dict[str, Any]):
    """A :mod:`pydantic` field type that serializes from URL-safe, base64-encoded
    input data.
    """
    __module__: str = 'ckms.core.models'

    @classmethod
    def __get_validators__(
        cls
    ) -> Generator[Callable[..., 'JSONB64'], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: bytes) -> 'JSONB64':
        if isinstance(value, cls): # pragma: no cover
            return value

        return cls(**json.loads(bytes.decode(b64decode(value), 'utf-8')))

    def __repr__(self) -> str: # pragma: no cover
        return f'JSONB64({super().__repr__()})'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast

from azure.keyvault.keys.crypto import SignatureAlgorithm

from ckms.types import Operation
from ckms.types import EllipticCurveOperation
from ckms.types import RSAOperation
from .models import AzureKeySpecification
from .types import AzureConsumer
from .types import IAzureProvider


class AzureProvider(IAzureProvider, AzureConsumer):
    __module__: str = 'ckms.ext.azure'
    spec: type[AzureKeySpecification] = AzureKeySpecification

    def get_operation_keyspec(self, op: Operation) -> AzureKeySpecification:
        return cast(AzureKeySpecification, op.get_keyspec())

    async def sign_azure(self, op: Operation) -> bytes:
        spec = self.get_operation_keyspec(op)
        async with self.crypto_client(spec.azure_key_id) as client:
            sig = await client.sign(
                algorithm=SignatureAlgorithm(spec.algorithm),
                digest=await op.get_digest()
            )
        return sig.signature

    async def sign_ec(self, op: EllipticCurveOperation) -> bytes:
        return await self.sign_azure(op)

    async def sign_rsa(self, op: RSAOperation) -> bytes:
        return await self.sign_azure(op)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio

import pytest
import pytest_asyncio

from ckms.core import Keychain
from ckms.core.models import KeySpecification


@pytest.fixture(scope='session')
def decrypter() -> Keychain:
    return Keychain()


@pytest.fixture(scope='session')
def encrypter() -> Keychain:
    return Keychain()


@pytest.fixture(scope='session')
def signer() -> Keychain:
    return Keychain()


@pytest.fixture(scope='session')
def verifier() -> Keychain:
    return Keychain()


@pytest_asyncio.fixture(scope='session', autouse=True) # type: ignore
async def setup_keys(
    decrypter: Keychain,
    encrypter: Keychain,
    signer: Keychain,
    verifier: Keychain,
    supported_keys: list[KeySpecification]
):
    await asyncio.gather(*supported_keys)
    for spec in supported_keys:
        assert spec.kid is not None
        assert spec.is_loaded()
        public = spec.as_public() if spec.is_asymmetric() else spec
        if spec.can_sign():
            signer.add(spec.kid, spec)
            verifier.add(public.kid, public) # type: ignore
        if spec.can_decrypt():
            decrypter.add(spec.kid, spec)
            encrypter.add(public.kid, public) # type: ignore


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Sized

import pytest

from ckms.core import Keychain
from ckms.core import const
from ckms.types import JSONWebKeySet


@pytest.fixture
def jwks(keychain: Keychain) -> JSONWebKeySet:
    return keychain.as_jwks(private=False)


def test_jwks_contains_asymmetric_keys(
    jwks: JSONWebKeySet,
    asymmetric_keys: Sized
):
    assert len(jwks) == len(asymmetric_keys)


@pytest.mark.parametrize("name", const.ASYMMETRIC_ALGORITHMS)
def test_jwks_has_key(
    keychain: Keychain,
    jwks: JSONWebKeySet,
    name: str
):
    spec = keychain.get(name)
    public = spec.as_public()
    assert jwks.has(public.kid)
    assert not jwks.has(name)


@pytest.mark.parametrize("name", const.ASYMMETRIC_ALGORITHMS)
def test_jwk_properties_match_key(
    keychain: Keychain,
    jwks: JSONWebKeySet,
    name: str
):
    spec = keychain.get(name)
    public = spec.as_public()
    jwk = jwks.get(public.kid)
    assert jwk.alg == public.algorithm
    assert jwk.crv == public.crv
    assert jwk.use == public.use
    assert set(jwk.key_ops) == public.key_ops


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Awaitable

from .ciphertext import CipherText
from .ioperationperformer import IOperationPerformer
from .iprovider import IProvider


class Decrypter(IOperationPerformer):
    __module__: str = 'ckms.types'
    algorithm: str | None
    provider: IProvider

    def decrypt(
        self,
        ct: bytes | CipherText,
        **kwargs: Any
    ) -> bytes | Awaitable[bytes]:
        """Decrypt `ct` using the configured algorithm."""
        assert hasattr(self, 'algorithm')
        assert hasattr(self, 'provider')
        assert isinstance(self.provider, IProvider)
        assert self.algorithm is not None
        if not isinstance(ct, CipherText): # pragma: no cover
            ct = CipherText(buf=ct)
        return self.provider.decrypt(
            algorithm=self.algorithm,
            key=self,
            ciphertext=ct
        )



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
from typing import Any

import pytest

from ckms.types import Digest


ALGORITHMS = [
    'sha256',
    'sha384',
    'sha512',
    'sha3_256',
    'sha3_384',
    'sha3_512'
]

HASHERS = [
    hashlib.sha256,
    hashlib.sha384,
    hashlib.sha512,
    hashlib.sha3_256,
    hashlib.sha3_384,
    hashlib.sha3_512,
]


class TestData:
    impl: type[Digest] = Digest

    @pytest.mark.parametrize("algorithm,factory", zip(ALGORITHMS, HASHERS))
    async def test_bytes(self, algorithm: str, factory: Any):
        digest = await self.impl.frombytes(algorithm, b'Hello world!')
        assert bytes(digest) == factory(b'Hello world!').digest()

    @pytest.mark.parametrize("algorithm,factory", zip(ALGORITHMS, HASHERS))
    async def test_digest(self, algorithm: str, factory: Any):
        digest = await self.impl.frombytes(algorithm, b'Hello world!')
        assert await digest.digest() == factory(b'Hello world!').digest()

    def test_digest_must_be_valid(self):
        with pytest.raises(ValueError):
            Digest(buf=b'foo', digest='foo')

    @pytest.mark.parametrize("algorithm,length", Digest.digest_size.items())
    def test_digest_must_be_valid_length(self, algorithm: str, length: int):
        with pytest.raises(ValueError):
            Digest(buf=b'0' * (length+1), digest=algorithm)

    @pytest.mark.asyncio
    async def test_mismatching_algorithm_raises_valueerror(self):
        digest = await self.impl.frombytes('sha256', b'Hello world!')
        with pytest.raises(ValueError):
            await digest.digest(algorithm='sha384')



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import ckms.core
from ckms.core.models import KeySpecification


SIGNING_KEYS_ASYMMETRIC: list[KeySpecification] = [
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'ec_sign_secp256k1_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'ec_sign_p256_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'ec_sign_p384_sha384',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pss_2048_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pss_3072_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pss_4096_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pss_4096_sha512',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pkcs1_2048_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pkcs1_3072_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pkcs1_4096_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_sign_pkcs1_4096_sha512',
        'version': 1
    }),
]

SIGNING_KEYS_SYMMETRIC: list[KeySpecification] = [
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'hmac_sha256',
        'version': 1
    })
]


ENCRYPTION_KEYS: list[KeySpecification] = [
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'aes256gcm',
        'version': 3
    }),

    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_2048_sha1',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_2048_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_3072_sha1',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_3072_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_4096_sha1',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_4096_sha256',
        'version': 1
    }),
    ckms.core.parse_spec({
        'provider': 'google',
        'project': 'unimatrixdev',
        'location': 'europe-west4',
        'keyring': 'local',
        'key': 'rsa_decrypt_oaep_4096_sha512',
        'version': 1
    }),
]

SUPPORTED_KEYS: list[KeySpecification] = (
    SIGNING_KEYS_ASYMMETRIC
    + SIGNING_KEYS_SYMMETRIC
    + ENCRYPTION_KEYS
)


@pytest.fixture(scope='session')
def supported_keys() -> list[KeySpecification]:
    return SUPPORTED_KEYS


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unimatrix.exceptions import CanonicalException


__all__: list[str] = [
    'AuthorizationServerMisbehaves',
    'AuthorizationServerNotDiscoverable',
    'TrustIssues',
    'UnresolvableIssuer',
    'UntrustedIssuer',
    'UntrustedSigningKey',
]


class TrustIssues(CanonicalException):
    """Base class for all exceptions related to trust."""
    __module__: str = 'ckms.types'
    http_status_code: int = 403
    code: str = 'TRUST_ISSUES'
    message: str = (
        "The application was not able to establish a trust relationship "
        "with the supplied credentials, authorization server, certificate "
        "authority or other third party."
    )


class UnresolvableIssuer(TrustIssues):
    """Raised when an :term:`Issuer` is not resolvable."""
    __module__: str = 'ckms.types'
    code: str = 'UNRESOLVABLE_ISSUER'


class AuthorizationServerNotDiscoverable(TrustIssues):
    """Raised when the application fails to discover the properties of an
    authorization server through standardized discovery protocols.
    """
    __module__: str = 'ckms.types'
    code: str = 'AUTHORIZATION_SERVER_NOT_DISCOVERED'
    detail: str = (
        "The metadata of the authorization server could not be discovered "
        "using well-known URLs or other standardized protocols."
    )


class AuthorizationServerMisbehaves(TrustIssues):
    """Raises when an authorization server does not behave as expected
    based on the protocol.
    """
    __module__: str = 'ckms.types'
    code: str = 'AUTHORIZATION_SERVER_MISBEHAVES'
    detail: str = (
        "The external authorization server does not behave according to its "
        "standardized specifications."
    )


class UntrustedIssuer(TrustIssues):
    """Raised when the issuer of a credential is not trusted by the
    application.
    """
    __module__: str = 'ckms.types'
    code: str = 'UNTRUSTED_ISSUER'


class UntrustedSigningKey(TrustIssues):
    """Raised when the signing key of a credential is not trusted by the
    application.
    """
    __module__: str = 'ckms.types'
    code: str = 'UNTRUSTED_SIGNING_KEY'


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import Any

import pydantic
from pyasn1.codec.ber import decoder
from pyasn1.type import univ
from pyasn1_modules import rfc2315


class ContentInfo(pydantic.BaseModel):
    content_type: str = pydantic.Field(..., alias='contentType')
    content: bytes

    @pydantic.validator('content_type', pre=True)
    def decode_content_type(cls, value: rfc2315.ContentType | str) -> str:
        if isinstance(value, rfc2315.ContentType):
            value = str(value)
        return value

    @pydantic.validator('content', pre=True)
    def decode_content(cls, value: bytes | univ.Any | univ.OctetString) -> bytes:
        if isinstance(value, univ.Any):
            value, _ = decoder.decode(bytes(value))
            if _:
                raise ValueError("Invalid BER encoding.")
            if not isinstance(value, univ.OctetString):
                raise TypeError(f'Invalid ASN.1 object: {type(value).__name__}')
        if isinstance(value, univ.OctetString):
            value = bytes(value)
        return bytes(value)


import asyncio

from ckms.jose.models import JSONWebKeySet


async def main():
    jwks = await JSONWebKeySet.discover("https://accounts.google.com")
    print(jwks)


if __name__ == '__main__':
    asyncio.run(main())


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ckms.ext import asn1


class SignatureDigestAlgorithmType(str, asn1.ObjectIdentifierEnum):
    # RFC 4055
    RSA_SHA256 = '1.2.840.113549.1.1.11'
    RSA_SHA384 = '1.2.840.113549.1.1.12'
    RSA_SHA512 = '1.2.840.113549.1.1.13'
    RSA_SHA224 = '1.2.840.113549.1.1.14'

    # RFC 5758
    ECDSA_SHA224 = '1.2.840.10045.4.3.1'
    ECDSA_SHA256 = '1.2.840.10045.4.3.2'
    ECDSA_SHA384 = '1.2.840.10045.4.3.3'
    ECDSA_SHA512 = '1.2.840.10045.4.3.4'

    def qualname(self) -> str:
        """Returns the default JWA-like algorithm."""
        alg = self.name[-3:]
        if str.startswith(self.name, 'RSA'):
            alg = f'RS{alg}'
        if str.startswith(self.name, 'ECDSA'):
            alg = f'ES{alg}'
        return alg


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.codec.der import decoder
from pyasn1.codec.der import encoder
from pyasn1.type import univ
from pyasn1_modules import rfc2315

from ckms.ext.x509.types import AlgorithmIdentifier
from .attribute import Attribute
from .issuerandserialnumber import IssuerAndSerialNumber


class SignerInfo(pydantic.BaseModel):
    version: int

    issuer: IssuerAndSerialNumber = pydantic.Field(
        default=...,
        alias='issuerAndSerialNumber'
    )

    digest_algorithm: AlgorithmIdentifier = pydantic.Field(
        default=...,
        alias='digestAlgorithm'
    )

    authenticated_attrs: list[rfc2315.Attribute] | None = pydantic.Field(
        default=None,
        alias='authenticatedAttributes'
    )

    encryption_algorithm: str = pydantic.Field(
        default=...,
        alias='digestEncryptionAlgorithm'
    )

    signature: bytes = pydantic.Field(
        default=...,
        alias='encryptedDigest'
    )

    unauthenticated_attrs: list[Attribute] | None = pydantic.Field(
        default=None,
        alias='unauthenticatedAttributes'
    )

    @pydantic.validator('digest_algorithm', pre=True)
    def validate_digest_algorithm(
        cls,
        value: rfc2315.DigestAlgorithmIdentifier | str | AlgorithmIdentifier
    ) -> str | AlgorithmIdentifier:
        if isinstance(value, rfc2315.DigestAlgorithmIdentifier):
            value = str(value['algorithm']) # type: ignore
        return value


    @pydantic.validator('encryption_algorithm', pre=True)
    def validate_encryption_algorithm(
        cls,
        value: rfc2315.DigestEncryptionAlgorithmIdentifier | str
    ) -> str:
        if isinstance(value, rfc2315.DigestEncryptionAlgorithmIdentifier):
            value = str(value['algorithm']) # type: ignore
        return value

    @pydantic.validator('signature', pre=True)
    def validate_signature(
        cls,
        value: bytes | rfc2315.EncryptedDigest
    ) -> bytes:
        if isinstance(value, rfc2315.EncryptedDigest):
            value = bytes(value)
        return value

    @pydantic.validator('authenticated_attrs', pre=True)
    def validate_authenticated_attrs(
        cls,
        value: list[rfc2315.Attribute] | rfc2315.Attributes | None
    ) -> list[rfc2315.Attribute] | None:
        if isinstance(value, rfc2315.Attributes):
            value = list(value)
        return value

    @pydantic.validator('unauthenticated_attrs', pre=True)
    def validate_unauthenticated_attrs(
        cls,
        value: list[rfc2315.Attribute] | rfc2315.Attributes | None
    ) -> list[rfc2315.Attribute] | None:
        if isinstance(value, rfc2315.Attributes):
            value = list(value)
        return value

    @property
    def signed_data(self) -> bytes:
        """The digest that was signed by the signer."""
        obj = univ.SetOf()
        for i, attr in enumerate(self.authenticated_attrs or []):
            obj.setComponentByPosition(i, attr) # type: ignore
        return self.get_digest(encoder.encode(obj)) # type: ignore

    @property
    def signed_digest(self) -> bytes:
        """The digest of the signed data."""
        digest: bytes
        for attr in (self.authenticated_attrs or []):
            t = attr[0]
            v = attr[1]
            if str(t) != '1.2.840.113549.1.9.4':
                continue
            digest, _ = decoder.decode( # type: ignore
                bytes(v[0]),
                asn1Spec=univ.OctetString()
            )
            if _:
                raise ValueError("Invalid DER encoding.")
            break
        else:
            raise ValueError("No digest present in SignerInfo.")
        return bytes(digest)

    def get_digest(self, data: bytes) -> bytes:
        """Return a byte-string containing a digest of `data` using the
        hashing algorithm specified by the signer.
        """
        return self.digest_algorithm.digest(data)

    def verify(
        self,
        key: ec.EllipticCurvePublicKey | rsa.RSAPublicKey,
        data: bytes
    ) -> bool:
        return False

    class Config:
        arbitrary_types_allowed: bool = True


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Literal

import pydantic
from cryptography.hazmat.primitives.ciphers import algorithms

from ckms.core import const
from ckms.types import IKeyInspector
from ckms.types import IProvider
from ckms.types import KeyUseType
from .contentencryptionkey import ContentEncryptionKey
from .hmac import HMAC
from .localkey import LocalKey
from .localkeyspecification import LocalKeySpecification
from .parameterlesskey import ParameterLessKey
from .transientkey import TransientKey


class SymmetricKeySpecification(LocalKeySpecification):
    kty: Literal['oct'] = 'oct'
    use: KeyUseType | None = None
    key: ContentEncryptionKey | LocalKey | TransientKey | ParameterLessKey | None

    @pydantic.root_validator(pre=True)
    def set_key_params(cls, values: dict[str, Any]) -> dict[str, Any]:
        algorithm = values.get('algorithm')
        if algorithm is not None and not bool(values.get('key')):
            # TODO: Very ugly
            if str.startswith(algorithm, 'A128'):
                length = 16
            elif str.startswith(algorithm, 'A192'):
                length = 24
            elif str.startswith(algorithm, 'A256'):
                length = 32
            else:
                raise NotImplementedError(algorithm)
            values['key'] = {'length': length}

        # If the key is specified by path, update the mapping with
        # encoding=binary to differentiate from PEM-encoded keys.
        key: dict[str, Any] | None = values.get('key')
        if key is not None:
            key['encoding'] = 'binary'
        return values

    @classmethod
    def autodiscover(
        cls,
        provider: IProvider,
        inspector: IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('algorithm') and not values.get('use'):
            raise ValueError(
                "Specify either the `algorithm` parameter or the "
                "`use` parameter."
            )
        if values.get('algorithm'):
            values['use'] = inspector.get_algorithm_use(values['algorithm'])
        elif values.get('use') == 'sig':
            values['algorithm'] = const.DEFAULT_SYMMETRIC_SIGNING_ALGORITHM
        elif values.get('use') == 'enc':
            values['algorithm'] = const.DEFAULT_SYMMETRIC_ENCRYPTION_ALGORITHM

    def get_private_bytes(self) -> bytes:
        private = self.get_private_key()
        return private.key

    def get_private_key(self) -> algorithms.AES | HMAC:
        assert isinstance(self.key, (algorithms.AES, HMAC)), type(self.key)
        return self.key

    async def setup_key(self) -> algorithms.AES | HMAC:
        k: bytes | algorithms.AES | HMAC = await super().setup_key()
        if isinstance(k, (algorithms.AES, HMAC)):
            return k
        assert isinstance(k, bytes), type(k) # nosec
        if self.use == 'sig':
            k = HMAC(k)
        elif self.use == 'enc':
            k = algorithms.AES(k)
        else:
            raise NotImplementedError(self.use)
        assert isinstance(k, (algorithms.AES, HMAC)) # nosec
        return k


"""Declares :class:`TrustedIssuers`."""
import asyncio
import warnings
from typing import Any
from typing import Generator

import httpx # TODO

from ckms.core import Keychain
from ckms.types import TrustIssues
from ckms.types import JSONWebKey
from ckms.types import JSONWebKeySet
from ckms.types import ServerMetadata


class TrustedIssuers:
    """Maintains a registry of trusted issuers and their signing/encryption
    keys.
    """
    deferred: object = object()
    issuers: dict[str, JSONWebKeySet]
    trusted: set[str]
    UntrustedIssuer: type[Exception] = type('UntrustedIssuer', (Exception,), {})

    @staticmethod
    def default() -> 'TrustedIssuers':
        """Return a :class:`TrustedIssuers` instance that is integrated
        with the settings module.
        """
        return _default

    @staticmethod
    def normalize_issuer(issuer: str) -> str:
        if not str.startswith(issuer, "https://"):
            issuer = f'https://{issuer}'
        return issuer

    def __init__(
        self,
        trusted: set[str] | None = None
    ):
        self.issuers = {}
        self.trusted = trusted or set()

    def get(self, issuer: str) -> JSONWebKeySet:
        """Return a :class:`JSONWebKeySet` holding the keys of the
        given `issuer`, or raise :exc:`ckms.types.TrustIssues`.
        """
        if issuer not in self.issuers:
            raise TrustIssues(
                detail=(
                    "The issuer of the credential attached to the request "
                    "is not trusted by the application."
                )
            )
        return self.issuers[issuer]

    def is_trusted(self, issuer: str | None) -> bool:
        """Return a boolean indicating if the given issuer is trusted."""
        return issuer in self.issuers

    def trust(
        self,
        issuer: str,
        keys: list[JSONWebKey] | JSONWebKeySet | Keychain | None = None
    ) -> None:
        """Add a trusted issuer to the registry.

        If the `keys` parameter is provided, no attempt is made to fetch
        the issuers' keys using any mechanism.
        """
        issuer = self.normalize_issuer(issuer)
        self.trusted.add(issuer)
        if keys is not None:
            if isinstance(keys, Keychain):
                keys = keys.as_jwks(private=False)
            elif isinstance(keys, list):
                keys = JSONWebKeySet(keys=keys)
            self.issuers[issuer] = keys

    async def setup(self) -> 'TrustedIssuers':
        """Fetch the remote keys for all issuers and setup the
        :class:`~ckms.types.JSONWebKeySet` instances.
        """
        async with httpx.AsyncClient(verify=False) as client:
            await asyncio.gather(*[
                self.setup_issuer(client, x)
                for x in self.trusted
            ])
        return self

    async def setup_issuer(
        self,
        client: httpx.AsyncClient,
        issuer: str,
        retry: int = 5,
        _attempts: int = 0
    ) -> None:
        """Setup the previously registered issuer `issuer`."""
        # Figure out the issuer URL. If it is a domain, prepend https. We assume
        # the OAuth 2.0 discovery protocol here.
        if str.startswith(issuer, 'http://'):
            raise ValueError("Can not fetch issuer keys over HTTP.")
        try:
            metadata = await ServerMetadata.discover(issuer=issuer, client=client)
            jwks = await metadata.get_jwks(client=client)
        except Exception:
            if _attempts >= retry:
                raise
            warnings.warn(f"Unable to reach issuer {issuer}")
            await asyncio.sleep(1)
            return await self.setup_issuer(
                client=client,
                issuer=issuer,
                retry=retry,
                _attempts=_attempts + 1
            )

        if jwks is None:
            raise ValueError(f"Unable to retrieve issuer JWKS: {issuer}")
        assert jwks is not None
        self.issuers[issuer] = jwks

    def verify_jws(self, jws: bytes | str) -> bool:
        """Verifies a JSON Web Signature (JWS) issued by one of the
        trusted issuers.
        """
        return False

    def __await__(self) -> Generator[Any, None, 'TrustedIssuers']:
        return self.setup().__await__()


_default: TrustedIssuers = TrustedIssuers()



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from typing import Any

import pytest

import ckms.core
from ckms.core.models import KeySpecification
from ckms.types import CipherText
from ckms.types import Message
from ckms.types import IProvider
from ckms.types import Verifier


class BaseProviderTestCase:
    name: str

    @property
    def provider(self) -> IProvider:
        """The :class:`~ckms.core.types.IProvider` implementation that
        is being tested.
        """
        return self.get_provider()

    @classmethod
    def setup_class(cls) -> None:
        pass

    def get_provider(self) -> IProvider:
        """Returns the :class:`~ckms.core.types.IProvider` instance that
        is being tested.
        """
        return ckms.core.provider(self.name)

    async def decrypt(
        self,
        spec: KeySpecification,
        ct: CipherText
    ) -> bytes:
        await spec
        pt = spec.decrypt(ct, force=True)
        if inspect.isawaitable(pt):
            pt = await pt
        return pt

    async def encrypt(
        self,
        spec: KeySpecification,
        pt: bytes,
        **kwargs: Any
    ) -> CipherText:
        await spec
        kwargs.setdefault('force', True)
        ct = spec.encrypt(pt, **kwargs)
        if inspect.isawaitable(ct):
            ct = await ct
        return ct

    async def sign(
        self,
        spec: KeySpecification,
        data: bytes | Message
    ) -> bytes:
        await spec
        sig = spec.sign(data)
        if inspect.isawaitable(sig):
            sig = await sig
        return sig

    async def verify(
        self,
        spec: Verifier,
        sig: bytes,
        data: bytes | Message
    ) -> bool:
        result = spec.verify(sig, data)
        if inspect.isawaitable(result):
            result = await result
        return result


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from cryptography.hazmat.primitives.asymmetric.types import PUBLIC_KEY_TYPES
from pyasn1_modules import rfc2986

from ckms.ext import asn1
from ckms.ext.rfc5280.types import RDNSequence
from ckms.ext.rfc5280.types import SubjectPublicKeyInfo
from ..oid import ID_EXTENSION_REQUEST
from .attributelist import AttributeList
from .certificateextensionsequence import CertificateExtensionSequence


class CertificationRequestInfo(asn1.ASN1Model):
    __asn1spec__: type[rfc2986.CertificationRequestInfo] = rfc2986.CertificationRequestInfo
    version: asn1.fields.Integer = asn1.fields.Integer(0)
    subject: RDNSequence = RDNSequence()
    subject_public_key: SubjectPublicKeyInfo = pydantic.Field(
        default=...,
        alias='subjectPKInfo'
    )
    attributes: AttributeList = AttributeList()
    _extensions: CertificateExtensionSequence = pydantic.PrivateAttr(
        default=CertificateExtensionSequence([])
    )

    @property
    def extensions(self) -> CertificateExtensionSequence:
        return self._extensions

    @classmethod
    def new(cls, key: PUBLIC_KEY_TYPES) -> 'CertificationRequestInfo':
        return cls(
            subjectPKInfo=SubjectPublicKeyInfo.frompublic(key)
        )

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if self.attributes.has(ID_EXTENSION_REQUEST):
            self._extensions = CertificateExtensionSequence.from_attribute(
                attribute=self.attributes.get(ID_EXTENSION_REQUEST)
            )

    def asn1(self) -> asn1.Asn1Type:
        self.attributes.add(self._extensions.as_attribute())
        return super().asn1()


import asyncio

import ckms.core


async def main():
    keys = [
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'aes256gcm',
            'version': 3
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'ec_sign_secp256k1_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'ec_sign_p256_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'ec_sign_p384_sha384',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_2048_sha1',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_2048_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_3072_sha1',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_3072_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_4096_sha1',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_4096_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_decrypt_oaep_4096_sha512',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pss_2048_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pss_3072_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pss_4096_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pss_4096_sha512',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pkcs1_2048_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pkcs1_3072_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pkcs1_4096_sha256',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'rsa_sign_pkcs1_4096_sha512',
            'version': 1
        }),
        ckms.core.parse_spec({
            'provider': 'google',
            'project': 'unimatrixdev',
            'location': 'europe-west4',
            'keyring': 'local',
            'key': 'hmac_sha256',
            'version': 1
        })
    ]
    await asyncio.gather(*keys)
    for k in keys:
        print(f'Key {k.kid}, {k.kty}, {k.algorithm}')


if __name__ == '__main__':
    asyncio.run(main())




# pylint: skip-file
import asyncio

import ckms
from ckms.jose.models import JSONWebKey
from ckms.jose import TrustedIssuers


async def main():
    external = ckms.Keychain()
    external.register({
        'sig': {
            'provider': "local",
            'kty': "OKP",
            'crv': "Ed25519",
        },
        'enc': {
            'provider': "local",
            'kty': "OKP",
            'crv': "X25519",
        },
    })
    await external.setup()

    issuers = TrustedIssuers()
    #issuers.trust("https://accounts.google.com")
    #issuers.trust("https://localhost:8000", [
    #    JSONWebKey.generate(kty='OKP', crv='Ed25519'),
    #    JSONWebKey.generate(kty='OKP', crv='Ed25519', kid='foo'),
    #])
    issuers.trust("https://localhost:8001", external)
    await issuers.setup()
    print(issuers.issuers)


if __name__ == '__main__':
    asyncio.run(main())



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import AsyncIterable

import pytest
import pytest_asyncio
from httpx import AsyncClient

from ckms.ext.tsa import Timestamper
from ckms.ext.tsa import TimestampingAuthority
from ckms.types import Digest
from ckms.types import Message


@pytest_asyncio.fixture # type: ignore
async def client() -> AsyncIterable[AsyncClient]:
    async with AsyncClient() as client:
        yield client


@pytest.fixture
def data() -> bytes:
    return b'Hello world!'


@pytest_asyncio.fixture # type: ignore
async def digest(message: Message) -> Digest:
    return Digest.frombytes('sha256', bytes(message))


@pytest.fixture
def message(data: bytes) -> Message:
    return Message(buf=data)


@pytest.fixture
def timestamper(tsa: TimestampingAuthority) -> Timestamper:
    return Timestamper([tsa])


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from ckms.core.models import KeySpecification
from ckms.utils import b64encode


class JOSEBaseTestCase:
    __module__: str = 'ckms.core.providers.tests'
    supports_compact: bool = True
    supports_flattened: bool = True
    supports_json: bool = True
    supports_multi: bool = True

    def jwk_from_spec(self, spec: KeySpecification, private: bool = False) -> dict[str, Any]:
        k: dict[str, Any]
        if spec.is_symmetric():
            k = dict(
                kty='oct',
                kid=spec.kid,
                k=b64encode(spec.get_key_material()).decode('ascii')
            )
        else:
            k = dict(**spec.as_jwk(private=private).dict(exclude_defaults=True))
        return k


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Encoder`."""
from typing import cast
from typing import Any

from ckms.types import ClaimSet
from ckms.types import Data
from ckms.types import IKeychain
from ckms.types import Encrypter
from ckms.types import Signer
from .josepayload import JOSEPayload


class Encoder:
    """Provides an interface to create JOSE objects."""
    __module__: str = 'ckms.jose'
    allow_compact: bool
    allow_flattened: bool
    encrypter: IKeychain
    encryption_keys: list[str]
    signer: IKeychain
    signing_keys: list[str | Signer]

    def __init__(
        self,
        *,
        encrypter: IKeychain,
        signer: IKeychain,
        encryption_keys: list[str] | None = None,
        signing_keys: list[str | Signer] | None = None,
        allow_compact: bool = True,
        allow_flattened: bool = True
    ) -> None:
        self.allow_compact = allow_compact
        self.allow_flattened = allow_flattened
        self.encrypter = encrypter
        self.encryption_keys = encryption_keys or []
        self.signer = signer
        self.signing_keys = signing_keys or []

    def encode(
        self,
        content: bytes | dict[str, Any] | str | ClaimSet | Data,
        sign: bool = True,
        encrypt: bool = True,
        content_type: str | None = None,
        claimset_class: type[ClaimSet] = ClaimSet,
        signing_keys: list[str] | None = None,
        signers: list[str | Signer] | list[str] | None = None,
        encrypters: list[str] | list[Encrypter] | None = None,
        allow_compact: bool = True,
        allow_flattened: bool = True
    ) -> JOSEPayload:
        """Encode claim set or readable byte-sequence `content` using the
        :class:`~ckms.types.IKeychain` supplied at intialziation.
        
        If `content` is a dictionary, assume that it is a JWT claims set. Otherwise,
        `content` is :class:`ckms.types.ClaimSet` or :class:`ckms.types.Data`. In the
        latter case, the instance *must* set the :attr:`~ckms.types.Data.content_type`
        attribute.

        For the case that a JWT claims set is being encoded but its content type
        is not ``JWT``, the `content_type` parameter may be supplied to override
        the default content type, for example when encoding an :rfc:`RFC9068`
        access token which uses ``at+jwt`` as its content type.
        """
        encrypters = encrypters or self.encryption_keys
        must_encrypt = encrypt and bool(encrypters)
        must_sign = sign or bool(signing_keys) or bool(signers)
        signers = signers or signing_keys or self.signing_keys
        if isinstance(content, str):
            content = str.encode(content, 'utf-8')
        if isinstance(content, bytes):
            content = Data(buf=content, content_type="application/octet-stream")
        if isinstance(content, dict):
            content = claimset_class(**content)
        obj = JOSEPayload(
            payload=content,
            content_type=content_type,
            allow_compact=self.allow_compact or allow_compact,
            allow_flattened=self.allow_flattened or allow_flattened
        )
        if must_sign:
            for signer in signers:
                if isinstance(signer, str):
                    signer = cast(Signer, self.signer.get(signer))
                self._sign_jws(obj, signer)
        if must_encrypt:
            for encrypter in encrypters:
                if isinstance(encrypter, str):
                    encrypter = cast(Encrypter, self.encrypter.get(encrypter))
                obj.add_recipient(
                    encrypter=encrypter,
                    direct=encrypter.can_encrypt()
                )
                
        return obj

    def _sign_jws(self, obj: JOSEPayload, signer: Signer) -> None:
        assert self.signer is not None
        obj.add_signature(
            signer=signer,
            protected={'alg': signer.algorithm, 'kid': signer.kid}
        )


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from ckms.types import CipherText
from ckms.types import Encrypter
from ckms.types import IContentEncryptionKey
from ckms.utils import b64encode_str


class Recipient:
    __module__: str = 'ckms.jose'
    direct: bool
    encrypter: Encrypter
    encryption_key: str
    header: dict[str, Any]

    def __init__(
        self,
        *,
        encrypter: Encrypter,
        header: dict[str, Any] | None = None,
    ):
        if encrypter.kid is None:
            raise ValueError("Encrypter.kid must not be None")
        self.encrypter = encrypter
        self.encryption_key = ''
        self.header = header or {}

    def dict(self, compact: bool = False, flatten: bool = False) -> dict[str, Any]:
        """Return a datastructure describing the recipient."""
        return {
            'encrypted_key': self.encryption_key,
            'header': self.header
        }

    async def encrypt(self, cek: IContentEncryptionKey) -> None:
        """Encrypt the Content Encryption Key (CEK) using the recipient
        key.
        """
        # If cek and self.encrypter are the same, then Direct Encryption or
        # Direct Key Agreement is used.
        is_direct = cek == self.encrypter
        self.header['kid'] = self.encrypter.kid
        if not is_direct:
            await cek
            ct = await self.encrypter.wrap(bytes(cek))
            assert isinstance(ct, CipherText)
            self.header['alg'] = self.encrypter.algorithm
            if ct.iv:
                self.header['iv'] = b64encode_str(ct.iv)
            if ct.tag:
                self.header['tag'] = b64encode_str(ct.tag)
            self.encryption_key = bytes.decode(ct.base64(), 'ascii')
        elif is_direct:
            self.header['alg'] = 'dir'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Awaitable

from .ciphertext import CipherText
from .ioperationperformer import IOperationPerformer
from .iprovider import IProvider
from .keyoperationtype import KeyOperationType
from .plaintext import PlainText


class Encrypter(IOperationPerformer):
    __module__: str = 'ckms.core.types'
    algorithm: str | None
    kid: Any
    provider: IProvider

    @IOperationPerformer.must_allow(KeyOperationType.encrypt)
    def encrypt(
        self,
        pt: bytes | PlainText,
        **kwargs: Any
    ) -> CipherText | Awaitable[CipherText]:
        """Encrypt `pt` using the configured algorithm."""
        assert hasattr(self, 'algorithm')
        assert hasattr(self, 'provider')
        assert isinstance(self.provider, IProvider)
        assert self.algorithm is not None
        if not isinstance(pt, PlainText):
            pt = PlainText(pt)
        return self.provider.encrypt(
            key=self,
            algorithm=self.algorithm,
            plaintext=pt,
            **kwargs
        )

    @IOperationPerformer.must_allow(KeyOperationType.wrapKey)
    def wrap(
        self,
        key: bytes | PlainText,
        **kwargs: Any
    ) -> CipherText | Awaitable[CipherText]:
        assert hasattr(self, 'algorithm')
        assert hasattr(self, 'provider')
        assert isinstance(self.provider, IProvider)
        assert self.algorithm is not None
        if not isinstance(key, PlainText):
            key = PlainText(key)
        return self.provider.wrap(
            key=self,
            algorithm=self.algorithm,
            wrap=key,
            **kwargs
        )

    def is_aead(self) -> bool:
        """Return a boolean indicating if the operator supports Authenticated
        Encryption with Associated Data (AEAD).
        """
        raise NotImplementedError


"""Declares :class:`JSONWebEncryption`."""
import typing

import pydantic

from .jsonwebencryptionrecipient import JSONWebEncryptionRecipient
from .jsonwebencryptionflattened import JSONWebEncryptionFlattened
from .jsonwebencryptionwithrecipients import JSONWebEncryptionWithRecipients
from .jsonb64 import JSONB64


class JSONWebEncryption(pydantic.BaseModel):
    __module__: str = 'ckms.core.models'
    __root__: typing.Union[
        JSONWebEncryptionFlattened,
        JSONWebEncryptionWithRecipients
    ]

    @property
    def header(self) -> dict[str, typing.Any]:
        """The JOSE header containing the metadata describing the
        encryption and the encrypted object.
        """
        return self.__root__.get_header()

    @property
    def recipients(self) -> list[JSONWebEncryptionRecipient]:
        return self.__root__.recipients

    @classmethod
    def parse_compact(cls, token: bytes) -> 'JSONWebEncryption':
        protected, encryption_key, iv, ciphertext, tag = token.split(b'.')
        return cls.parse_obj({
            'protected': protected,
            'unprotected': {},
            'header': {},
            'encrypted_key': encryption_key,
            'iv': iv,
            'ciphertext': ciphertext,
            'tag': tag,
            'aad': None
        })

    @pydantic.root_validator(allow_reuse=True)
    def validate_protected(cls, values: dict[str, typing.Any]):
        dto = typing.cast(
            typing.Union[
                JSONWebEncryptionFlattened,
                JSONWebEncryptionWithRecipients
            ],
            values.get('__root__')
        )
        protected = typing.cast(JSONB64, dto.protected or {})
        unprotected = dto.unprotected

        # The Header Parameter values used when creating or validating
        # per-recipient ciphertext and Authentication Tag values are
        # the union of the three sets of Header Parameter values that
        # may be present: (1) the JWE Protected Header represented in
        # the "protected" member, (2) the JWE Shared Unprotected Header
        # represented in the "unprotected" member, and (3) the JWE
        # Per-Recipient Unprotected Header represented in the "header"
        # member of the recipient's array element.  The union of these
        # sets of Header Parameters comprises the JOSE Header.  The
        # Header Parameter names in the three locations MUST be
        # disjoint (RFC 7516, Section 7.2.1).
        if set(protected.keys()) & set(unprotected.keys()):
            raise ValueError(
                "Parameters in protected and unprotected headers "
                "must be disjoint."
            )

        headers = {**protected, **unprotected}
        if isinstance(dto, JSONWebEncryptionFlattened):
            if bool(set(headers.keys()) & set(dto.header.keys())):
                raise ValueError(
                    "Parameters in protected, unprotected and recipient "
                    "headers must be disjoint."
                )

        for i, recipient in enumerate((dto.recipients)):
            if isinstance(dto, JSONWebEncryptionFlattened)\
            or not bool(set(recipient.header.keys()) & set(headers.keys())):
                continue
            raise ValueError(
                "Parameters in protected, unprotected and recipient "
                f"headers must be disjoint. Recipient {i} specified "
                "a header that was already in the protected or "
                "unprotected headers."
            )

        return values


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Awaitable

from .digest import Digest
from .keyoperationtype import KeyOperationType
from .ioperationperformer import IOperationPerformer
from .iprovider import IProvider
from .message import Message


class Signer(IOperationPerformer):
    __module__: str = 'ckms.types'
    algorithm: str | None
    provider: IProvider
    kid: Any

    @IOperationPerformer.must_allow(KeyOperationType.sign)
    def sign(
        self,
        message: bytes | Digest | Message,
    ) -> bytes | Awaitable[bytes]:
        """Sign `message` using the configured algorithm."""
        assert hasattr(self, 'algorithm')
        assert hasattr(self, 'provider')
        assert isinstance(self.provider, IProvider)
        assert self.algorithm is not None
        digest = self.provider.get_hashing_algorithm(self.algorithm)
        if not isinstance(message, (Digest, Message)):
            message = Message(buf=message)
        if digest is not None and not message.digestmod:
            message.set_digest_algorithm(digest)
        return self.provider.sign(
            key=self,
            algorithm=self.algorithm,
            message=message
        )

    def get_digest_oid(self) -> str | None:
        """Return the OID of the signing algorithm."""
        raise NotImplementedError